! Module for general file processing operations
! Author: John Manchuk, 2010
! Modified: Ryan Barnett, 2012: added several subroutines (specified below)
!
! NOTES:
!   1. The file_open function has been written to return a valid
!      file handle so hard coding a handle value (ex. lin = 1, lout = 2)
!      is not required.  The function determines a valid handle and
!      returns it to the calling program.
!   2. Reading through a GSLIB style text file is straightforward using
!      the "datafilesize" and "readfileheader" routines.  The file does
!      not have to be pre-opened to call these.  The following is
!      recommended to read in a GSLIB file:
!           call datafilesize ( fid, fname, ncol, nrow )
!           ! allocate a real array ncol by nrow, ex.
!           allocate(filedata(ncol,nrow))
!           ncol = readfileheader ( fid, fname )
!           read(fid,*) filedata
!

module filehandling
    use ifport
    implicit none
    
    public :: file_exists, &        !check if a file exists
              file_open, &          !open a file and return the handle (integer function)
              readfileheader, &     !read through a gslib file header and return number of columns (integer function)
              datafilesize,&        !determine the number of rows and columns in a file
              findpar,&             !find a parameter file and the start of its 'interesting' lines
              checkname             !for handling file names
    private
    
    contains
    
    !Determine if a file exists
    logical function file_exists ( fname )
        character*(*), intent (in) :: fname
        integer :: does_exist
        file_exists = .false.
        inquire ( file = fname, exist = does_exist )
        if(does_exist /= 0) file_exists = .true.
    end function file_exists
    
    !Open a file
    ! This returns the handle value
    ! A handle of -1 indicates an error
    integer function file_open ( fname, create, access_ ) result ( fid )
        character*(*), intent (in) :: fname
        logical, intent (in) :: create
        character*(*), optional :: access_
        logical :: isopen
        integer :: fidt
        if (not(file_exists ( fname )) .and. not(create) )then
            write(*,*) 'Error! Datafile does not exist:',fname(1:30)
        endif
        !get next available link
        fidt = 0
        isopen = .true.
        do while ( isopen )
            fidt = fidt + 1
            if(fidt == 6) fidt = fidt + 1 !skip command window handle
            inquire( fidt, OPENED = isopen )
        enddo
        
        fid = -1
        if(create)then
            fid = fidt
            if(present(access_))then
                open ( fid, file = fname, status = 'unknown', access = access_ )
            else
                open ( fid, file = fname, status = 'unknown' )
            endif
        else
            if( .not. file_exists ( fname ) ) return
            fid = fidt
            if(present(access_))then
                open ( fid, file = fname, status = 'old', access = access_ )
            else
                open ( fid, file = fname, status = 'old' )
            endif
        endif
        return
    end function file_open
    
    !Read through a GSLIB file header
    integer function readfileheader ( fid, fname ) result ( ncol )
        integer, intent (inout) :: fid
        character*(*), intent (in) :: fname
        logical :: isopen
        character*64 :: str
        integer :: i
        
        inquire(fid,OPENED = isopen)
        if(.not.isopen)then
            fid = file_open ( fname, .false. )
            if(fid < 0) return
        else
            rewind(fid)
        endif
        
        read(fid,'(a)',end=10,err=11) str
        read(fid,*,end=10,err=11) ncol
        do i = 1, ncol
            read(fid,'(a)',end=10,err=11) str
        enddo
        
        return
        
    10  write(*,'(a,a)') 'ERROR, END OF FILE: ',trim(fname)
        stop
    11  write(*,'(a,a)') 'ERROR READING FILE: ',trim(fname)
        stop
    end function readfileheader
    
    !Determine the size of a data file with the typical row-by-row method
    subroutine datafilesize ( fid, fname, ncol, nrow )
        integer, intent (inout) :: fid !file handle
        character*(*), intent (in) :: fname
        integer, intent (out) :: ncol
        integer*4, intent (out) :: nrow
        logical :: isopen
        real :: rec
        
        inquire(fid,OPENED = isopen)
        if(.not.isopen)then
            fid = file_open ( fname, .false. )
            if(fid < 0) return
        else
            rewind(fid)
        endif
                
        ncol = readfileheader ( fid, fname )
        nrow = 0
        
        do
            read(fid,*,end=9,err=11) rec
            nrow = nrow + 1
        enddo
    9   rewind(fid)
        
        return
    10  write(*,'(a,a)') 'ERROR: END OF FILE: ',trim(fname)
        stop
    11  write(*,'(a,a)') 'ERROR READING FILE: ',trim(fname)
        stop
    end subroutine datafilesize

    ! Read a parameter file up to starting parameters 
    ! and consider a makepar call
    subroutine findpar(parname,ipar,lin)
        character*(*) :: parname
        character str*512
        integer ipar,lostr,lin
        
        lostr=len_trim(parname)
        str = ' '
        
        call getarg(1,str)
        if(str(1:1) == ' ')then
            write(*,*) 'Which parameter file do you want to use?'
            read (*,'(a)') str
        endif
        
        if(str(1:1) == ' ') str(1:lostr) = parname
        if(.not.file_exists(trim(str)))then
        write(*,*) 'ERROR - the parameter file does not exist,'
        write(*,*) '        check for the file and try again  '
        write(*,*)
        if(str(1:lostr) == parname) then
            if(ipar.eq.1)then
                write(*,*) '        creating a blank parameter file'
                call makepar
            else
                write(*,*) '        find a valid parameter file'
                stop
            endif
            write(*,*)
        endif
        stop
        endif
!
! Open the parameter file and find Start of Parameters:
!
        lin = file_open( trim(str), .false. )
        do while ( str(1:4) /= 'STAR' )
        read(lin,'(a4)',end=98) str(1:4)
        enddo
        
        return
        
 98     stop ' ERROR in parameter file'
        
    end subroutine findpar
  
   
        ! This subroutine takes the character string "str" of length "len" and
    ! removes all leading blanks and blanks out all characters after the
    ! first blank found in the string (leading blanks are removed first).
    subroutine checkname(temp)
        character*(*) :: temp
        integer       :: i

        ! Remove leading blanks:
        temp = adjustl(temp)
        
        ! find first two blanks and blank out remaining characters:
        i = index(temp,'  ')  ; if (i > 0) temp(i:)=' '

        ! Look for "-fi"
        i = index(temp,'-fi') ; if (i > 0) temp(i:)=' '

        ! Look for "\fi"
        i = index(temp,'\fi') ; if (i > 0) temp(i:)=' '
        return
    end subroutine checkname    


end module filehandling