
!***********************************************************************************!
!                                                                                   !
!  PROGRAM: PPMT                                                                    !
!                                                                                   !
!  PURPOSE:  Projection Pursuit Multivariate Transform                              !
!                                                                                   !
!  AUTHOR: Ryan Barnett and John Manchuk 2015                                       !
!                                                                                   !
!  PAPER: Barnett, R., Manchuk, J., & Deutsch, C. (2012). Projection Pursuit        !
!  Multivariate Transform. CCG Annual Report 14, Paper103.                          !
!                                                                                   !
!  ACKNOWLEDGMENTS: Based largely on the algorithm originally outlined in the paper:!
!  Friedman, J. (1987). Exploratory Projection Pursuit. Journal of the American     !
!  Statistical Association, vol.82, pp.249-266.                                     ! 
!                                                                                   !
!  Version 3.1 performs an optional spatial decorrelation (MAF) transform following ! 
!              projection pursuit                                                   !
!***********************************************************************************!

    program ppmt
    
        use geostat !Common variables
        use filehandling !Common data reading/writing functions
        use quicksort !Sorting module
        use ifport
        use normaldist, only: snorm_inv
        use ppmt_mod
        implicit none
        integer :: miniters, maxiters, nboot, iters, inner_iter, ptarget, ltrn, lit, exc, ilag, wtcol, &
            lout1
        real*8 :: a1, step, maxpix, maxpix0, pix, c1, c2, s, pixp, pixm, vec1, vec2, snpixm
        real*8, allocatable, dimension(:) :: indexi, snorm, snpixp
        real*8, allocatable, dimension(:,:) :: R, ei, a, id, U, Ux, eit, atest1, atest2, c, gradtest, &
            eye, Am, pr, w, z, y, refs, x, snpix, sph, outd, cov, data1, y1
        real*8,allocatable,dimension(:,:,:)::rsnorm
        character*512 :: trnfl, str1, outfl1
        character*5, parameter :: version = '3.201'
        character*12, parameter :: parname = 'ppmt.par'
        
        ! Varcalc parameters
        logical :: strict, usespatial, useweight
        integer*4 :: nvars, nvarg, nlags, xyzcols(3), is, nc, spatialiter, iter1, iters_type(10000), &
            nt, nvarios, itertype, len, ispatial
        real*8 :: azm, dip, tilt, diptol, lagtol, azmtol, bandhorz, bandvert, lagdist, inc  
        real*8, allocatable, dimension(:,:) :: xyz, variocutcat      
        logical, allocatable, dimension(:) :: variostd
        integer*4, allocatable, dimension(:) :: invariotypes, varioheads, variotails, varidx, varnumpairs  
        real*8, allocatable, dimension(:) :: variosills, varazm,vardip, varlagdist, varvalue, lagdists,&
            cov_spat_tot, lagindex   
        real*8, allocatable, dimension(:,:,:) :: cov_spat
!
! Note version and set fixed parameters
!   
        write(*,'(/,a,a,/)') ' PPMT version: ',version 
!
! Get the name of parameter file, open the file and find start of parameters
!
        call findpar(parname,1,lin)
!
! Read Input Parameters:
!
        read(lin,'(a512)',err=98) datafl
        call checkname(datafl)
        write(*,'(a,a)') ' input data file = ', trim(datafl)
        
        vcols=0
        read(lin,*,err=98) nvar, vcols(1:nvar), wtcol
        write(*,'(a,i4)') ' number of variables = ',nvar
        write(*,'(a,<(nvar+1)>(x,i4))') ' columns for variables and weight = ', vcols(1:nvar), wtcol
        
        if( wtcol < 1 )then
            useweight = .false.
            write(*,'(a)') ' using weights for transform'
        else
            useweight = .true.
            write(*,'(a)') ' not using weights for transform'
        endif

        read(lin,*,err=98) tmin, tmax
        write(*,'(a,2(x,e8.2))') ' trimming limits = ', tmin, tmax

        read(lin,*,err=98) miniters, maxiters, ptarget
        write(*,'(a,3(x,i4))') ' min/max iters. & targeted percentile = ', miniters, maxiters, ptarget
        
        is = 0
        read(lin,*,err=98) is
        if( is > 0 )then
            usespatial = .true.
            write(*,'(a,i4)') ' performing spatial decorrelation'
            read(lin,*,err=98) xyzcols
            write(*,'(a,<(3)>(x,i4))') ' xyz columns = ', xyzcols
            if( any( xyzcols(1:2) < 1 ) )then
                write(*,'(a)') ' Error: require x and y columns if using spatial decorrealtion!'
                write(*,'(a)') '        Only the z column is optional!'  
                stop
            endif
            if( xyzcols(3) > 0 )then
                nc = 3
            else
                nc = 2
            endif
            read(lin,*,err=98) lagdist, lagtol
            write(*,'(a,i4,2(x,f8.1),i4)') ' lagdist, lagtol = ', lagdist, lagtol
        else
            usespatial = .false.
            nc = 0
            write(*,'(a,i4)') ' not performing spatial decorrelation'
            read(lin,*,err=98); read(lin,*,err=98)
        endif
        
        read(lin,'(a512)',err=98) outfl1
        call checkname(outfl1)
        write(*,'(a,a)') ' output file with univariate Gaussian variables = ', trim(outfl1)
        
        read(lin,'(a512)',err=98) outfl
        call checkname(outfl)
        write(*,'(a,a)') ' output file with multivariate Gaussian variables = ', trim(outfl)
        
        read(lin,'(a512)',err=98) trnfl
        call checkname(trnfl)
        write(*,'(a,a)') ' output file with transform info = ', trim(trnfl)    
        close(lin)
!
! Load the data 
!
        write(*,*)      
        write(*,'(a,a)') ' loading the data'
        
        call datafilesize ( lin, datafl, ncol, nrow )
        ! Run some checks on the columns
        if( any( vcols(1:nvar) > ncol ) .or. any( xyzcols(1:nc) > ncol ) .or. wtcol > ncol )then
            write(*,'(a)') ' Error: specified column exceeds the number of data file columns!'
            stop
        endif
        allocate( data_(ncol,nrow), oindex(nrow), stat = test)
        if(test /= 0)then; write(*,'(/,a,/)') 'ERROR: ALLOCATION FAILED';stop;endif
        ncol = readfileheader( lin, datafl ) 
        read(lin,*,err=99) data_
        close(lin)
        ! Can only use homotopic observations with greater than 0 weight
        ndata = 0
        do i = 1, nrow
            if( any( data_(vcols(1:nvar),i) < tmin ) ) cycle
            if( any( data_(vcols(1:nvar),i) > tmax ) ) cycle
            if( useweight )then
                if( data_(wtcol,i) <= 0 ) cycle
            endif
            ndata = ndata + 1
            oindex(ndata) = i
        enddo
        if(ndata == 0 )then
            write(*,*) ' ERROR: all of the data was trimmed or had <= 0 weight!' 
            stop
        endif  
        ! Allocate and populate some working arrays
        allocate( z(ndata,nvar), y(ndata,nvar), y1(ndata,nvar), x(ndata,nvar), w(ndata,2),&
        & snorm(ndata), sph(nvar,nvar), Am(nvar,nvar), xyz(ndata,3),  stat = test )
        if(test /= 0)then; write(*,'(/,a,/)') 'ERROR: ALLOCATION FAILED';stop;endif
        z = transpose(data_(vcols(1:nvar),oindex(1:ndata)) )
        w = 1.d0
        if( useweight ) w(:,2) = data_( wtcol, oindex(1:ndata) )
        if( usespatial )then
            xyz = 0.d0
            xyz(:,1:nc) = transpose( data_( xyzcols(1:nc), oindex(1:ndata) ) )
        endif
        deallocate(data_)
!   
! Open the transformation file to begin recording the transform
!    
        ltrn = file_open( trnfl, .true., 'stream' )
        if( ltrn < 0 )then
            write(*,*) ' Error - could not open output transformation table file'
            stop
        endif
        str = 'PPMT transformation table file'; len = len_trim(str)
        write(ltrn) len; write(ltrn) str(1:len)
        write(ltrn) nvar
        write(ltrn) ndata
!
! Normal score preprocessing (always equally weighted)
!          
        write(*,*)      
        write(*,'(a,a)') ' normal scoring the data'
        w(:,1) = 1 ! No normal score weights are applied prior to the PPMT (applied after if necessary) 
        call nscore( z , w(:,1), y )
        y1 = y ! Store the univariate Gaussian data for writing out later
        ! These standard normal Gaussian values will be referenced later for Gaussianizing
        snorm = y(:,1)
        call qsortem( snorm , 1 , ndata )
        do i = 1, nvar
            call qsortem( z(:,i) , 1 , ndata )
        enddo
        write(ltrn) snorm
        write(ltrn) z !ndata x nvar
        deallocate(z)  
!
! Sphere the data
!      
        write(*,*)      
        write(*,'(a,a)') ' sphereing the data' 
        call sphere( y  , x , sph )  
        ! Write out the inverse of the Sphere matrix (needed for back-transform rather than Sphere)
        call invert ( sph , nvar )
        write(ltrn) sph
        if( not(usespatial) ) deallocate(sph)       
!
! Initialize looping variables
!
        allocate(R(nvar,nvar),ei(nvar,nvar),a(1,nvar),atest1(1,nvar),atest2(1,nvar),id(nvar,nvar),&
        & gradtest(1,nvar),U(nvar,nvar),Ux(ndata,nvar),indexi(ndata),r8temp2(ndata,1),&
        & eit(nvar,1),c(1,1),pr(ndata,1), eye(1,nvar), stat = test)
        if(test /= 0)then; write(*,'(/,a,/)') 'ERROR: ALLOCATION FAILED';stop;endif
        u=0.d0;id=0.d0;ei=0.d0;
        do i = 1, nvar
            id(i,i) = 1
            ei(i,i) = 1
        enddo
        ! A random basis will be used intially for R
        ! pseudo-random number generator is good enough for this purpose
        do i=1,nvar
            do j=1,nvar
                R(i,j)=rand(0)
            enddo
        enddo
        call orthogs(R) 
        do i = 1, ndata
            indexi(i) = i
        enddo  
!
! Establish targeted projection index
!        
        if( ptarget > 0 )then
            !Hardcode the number of bootstrap samples for now...
            nboot=100
            allocate( rsnorm( nboot,ndata,nvar ), snpix(nboot,nvar), snpixp(nboot) , stat = test)
            if(test /= 0)then; write(*,'(/,a,/)') 'ERROR: ALLOCATION FAILED';stop;endif
            do i = 1, nboot
                do j = 1, ndata
                    do k = 1, nvar
                        rsnorm(i,j,k) = snorm_inv( dble(rand(0)) ) 
                    enddo
                 enddo
                ! Random Gaussian distribution assembled
                ! Now, test the projection index along random orthogonal directions
                eye = 0.d0
                do j=1,nvar
                    eye(1,j) = 1
                    snpix(i,j) = projindex( eye, dble(rsnorm(i,:,:)), w(:,1), 10 )
                    eye = 0.d0
                enddo
            enddo
            snpixm = 0.d0
            snpixp = 0.d0
            do i = 1, nvar
                call qsortem( snpix(:,i) , 1 , nboot )
                snpixm = snpixm+snpix( ptarget ,i )
                snpixp = snpixp + snpix( : ,i )
            enddo
            snpixm = snpixm / dble(nvar)
            snpixp = snpixp / dble(nvar)
            deallocate(snpix,rsnorm)
        endif                      
!
! Main projection pursuit loop
!             
        write(*,*)      
        write(*,'(a,a)') ' performing projection pursuit iterations' 
        iters = 1
        exc = 0
        iter1 = 0
        
        nt = 0
        ispatial = 0
        iters_type = 0
        do while (iters <= maxiters) 
            nt = nt + 1
            ispatial = ispatial + 1

            ! First determine the maximum direction along the coordinate axes
            ! Since X is sphered, the coordinate axes are the principal components
            ! of the original data Y
            maxpix = 0.d0
            do i = 1,nvar
                eye(1,:) = ei(i,:)
                pix = projindex( eye, x, w(:,1), 10 )
                if( abs(pix) > maxpix )then
                    a(1,:) = ei(i,:)
                    maxpix = abs(pix)
                endif
            enddo         
            ! Now perform a course stepping around the most interesting coordinate
            ! axes according to Freidman '85 (pg256)
            maxpix0 = 0.d0
            inner_iter = 0;
            do while ( (maxpix-maxpix0) >= 1.0d-20 .and. inner_iter < 100 )
                maxpix0 = maxpix
                do j = 1,nvar
                    eit(:,1) = ei(j,:)
                    c = (1.d0+ matmul( a,eit ))**0.5d0
                    c1 = c(1,1) 
                    atest1(1,:) = 1/sqrt(2.0)*( a(1,:)+ei(j,:) ) / c1
                    pixp = projindex( atest1, x, w(:,1), 10)
                    c = (1.d0- matmul( a,eit ))**0.5d0
                    c2 = c(1,1) 
                    atest2(1,:) = 1/sqrt(2.0)*( a(1,:)-ei(j,:) )/ c2
                    pixm = projindex( atest2, x, w(:,1), 10 )
                    if( isnan( pixp ) .or. any(isnan( atest1 )) ) pixp = 0.d0
                    if( isnan( pixm ) .or. any(isnan( atest2 )) ) pixm = 0.d0 
                    if( pixp == 0.d0 .and. pixm == 0.d0 ) cycle                   
                    if( pixp> pixm )then
                        pix = pixp;s = 1.d0 !ones(1,length(a));
                    else
                        pix = pixm;s = -1.d0 !*ones(1,length(a));
                    endif
                    if( pix > maxpix )then
                        a(1,:) = 1/sqrt(2.d0)*(a(1,:)+s*ei(j,:)) / (1+s*a(1,j))**0.5 
                        maxpix = pix
                    endif
                enddo
                inner_iter = inner_iter + 1
            enddo
            
            ! Now apply gradient descent to maximize on a smaller scale
            step = 0.1d0
            inner_iter = 0
            maxpix0 = 0.d0
            gradtest = 0.d0
            do while ( (maxpix-maxpix0) >= 1.0d-20 .and. inner_iter < 100 )
                maxpix0 = maxpix
                !Apply gradient descent
                call gradient( a, x, w(:,1), 10, gradtest )
                atest1=a+step*gradtest
                ! Update objective
                pix = projindex( atest1, x, w(:,1), 10 )
                if( pix >  maxpix )then
                    a = a + step*gradtest
                    maxpix = pix
                endif
                inner_iter = inner_iter + 1
            enddo
            !On occasion, projection of full zeros passes optimization
            !The following hotfix fixes the anomaly... must be better understood                             
            if(maxpix > 1)then
                exc = exc + 1
                if( exc > nvar ) exc = 1
                a = 0.d0
                a(1,exc) = 1.d0
            else
                write(*,'(a,f8.6)') ' maximum projection index:',  maxpix
            endif

            call projection( a, x, pr )
            
            !Gaussianize along the projection
            U = R
            U(:,1) = a(1,:)
            call orthogs(U)!orthogonolize with Gram-Schmidt
            Ux = matmul(x,U)
            r8temp2(:,1) = indexi
            call qsortem( pr(:,1), 1, ndata, r8temp2(:,1) )
            Ux(int( r8temp2(:,1) ),1) = snorm !nscore along projection
            x = matmul(Ux,transpose(U))   !unproject
         
            !Write information to the transformation table
            itertype = 54321 ! Code for projection pursuit
            write(ltrn) itertype
            write(ltrn) U
            write(ltrn) pr(:,1)
 
            ! Stopping criteria
            if( ptarget > 0 )then
                if( snpixm >= maxpix .and. iters >= miniters )exit
            endif
            iters = iters + 1 
        enddo
!
! Iterations complete, report stopping criteria
!        
        if( snpixm >= maxpix .and. ptarget > 0 )then 
            write(*,*)
            write(*,'(a,i3)') ' projection pursuit finished based on targeted Gaussianity perc.:',ptarget
        elseif( ptarget > 0 )then 
            j = 0
            do i = 1,nboot
                if( maxpix > snpixp(i) ) j = j + 1
            enddo  
            write(*,*)
            write(*,'(a,i3)') ' projection pursuit finished based on max iters. Final Gaussianity perc.:',j
        endif
        if( usespatial )then
            ! The first step of MAF is to sphere the data at zero lag         
            y = x
            call sphere( y  , x , sph )  
        
            ! Allocate some working arrays and specify some standard parameters for the variogram calculation
            nvarg = nvar*(nvar-1)/2+nvar
            nlags = 2
            allocate( invariotypes(nvarg), variostd(nvarg), variosills(nvarg),&
            & varioheads(nvarg), variotails(nvarg), lagdists(maxiters), variocutcat(nvarg,2),&
            & varidx(nvarg*nlags), varlagdist(nvarg*nlags), varnumpairs(nvarg*nlags), varvalue(nvarg*nlags),&
            & varazm(nvarg*nlags), vardip(nvarg*nlags), cov(nvar,nvar), stat = test)
            if(test /= 0)then; write(*,'(/,a,/)') 'ERROR: ALLOCATION FAILED';stop;endif
            
            ! Jared Deutsch's varcalc routines will be used for calculating the variogram and cross-variogram
            ! The below parameters are required for these routines 
            strict = .true.
            azm = 0.d0
            dip = 0.d0
            tilt = 0.d0
            azmtol = 90.d0
            diptol = 90.d0
            bandhorz = 1.0d21
            bandvert = 1.0d21
            variocutcat = 0
            k = 0
            do i = 1,nvar
                do j = i,nvar
                    k = k + 1
                    varioheads(k) = i
                    variotails(k) = j
                    if( i ==j )then
                        invariotypes(k) = 1 ! Semivariogram
                    else
                        invariotypes(k) = 2 ! Cross-semivariogram
                        variostd = .false. ! Don't need to standardize
                        variosills = 0 ! Not used since not standardizing
                    endif
                enddo
            enddo
            ! Calculate the semivariogram and cross-semivariogram
            call varcalc( xyz, x, ndata, nvar, tmin, tmax, strict, &
            & invariotypes, variostd, variosills, varioheads, variotails, nvarg, &
            & azm, dip, tilt, azmtol, diptol, 1, lagdist, lagtol, bandhorz, bandvert, &
            & variocutcat, varidx, varlagdist, varnumpairs, varvalue, varazm, vardip)      
            
            ! Translate to a covariance matrix
            k = 0
            do i = 1,nvar
                do j = i,nvar
                    k = k + 1
                    cov(i,j) = varvalue(2+(k-1)*nlags)
                    cov(j,i) = cov(i,j)
                enddo
            enddo
            
            ! Decorrelate at this lag
            y = x
            call decorrelate( y, cov, x, Am)
            
            ! Calculate the MAF 'A' matrix for back-transforming
            Am =  matmul( sph, Am )
            ! Write out the inverse of the Sphere matrix (needed for back-transform rather than A)
            call invert( Am , nvar )
            itertype = 12345 ! Code for spatial decorrelation
            write(ltrn) itertype
            write(ltrn) Am        
        endif 
!
! Normalize the margins - apply declustering weights if necessary
!               
        itertype = 6789 ! Code for weighted transform
        write(ltrn) itertype
        if( useweight )then
            call nscore( x , w(:,2) , y )
        else
            call nscore( x , w(:,1) , y )
        endif  
        write(ltrn) x
        write(ltrn) y
        close(ltrn)
        x = y  
!
! Write the output data file
!
        write(*,*)      
        write(*,'(a,a)') ' writing the results' 
        
        allocate( vnames(nvar), data_(nrow,nvar), data1(nrow,nvar) )
        lout = file_open ( outfl, .true. )
        lout1 = file_open ( outfl1, .true. )
        lin = file_open ( datafl, .false. )
        read(lin,*)
        read(lin,*) nvari
        data_ = -999.0
        data_( oindex(1:ndata), : ) = x( 1:ndata, : )
        write(lout,'(a)') 'Input data & PPMT transformed variables'
        write(lout,'(i3)') nvari + nvar
        data1( oindex(1:ndata), : ) = y1( 1:ndata, : )
        write(lout1,'(a)') 'Input data & nscore transformed variables'
        write(lout1,'(i3)') nvari + nvar
        do i = 1, nvari
            read(lin,'(a512)') str
            write(lout,'(a)') trim(adjustl(str))
            write(lout1,'(a)') trim(adjustl(str))
            do j = 1, nvar
                if( i == vcols(j) )then
                    vnames(j) = trim(adjustl(str))
                endif
            enddo
        enddo
        do i = 1, nvar
            write(lout,'(a)') 'PPMT:'//trim(adjustl(vnames(i)))
            write(lout1,'(a)') 'Nscore:'//trim(adjustl(vnames(i)))
        enddo
        do i = 1, nrow
            read(lin,'(a512)') str
            write(lout,'(a)',advance='no') trim(str)
            write(lout,'(<(nvar)>(x,g15.8))') data_( i, 1:nvar )
            write(lout1,'(a)',advance='no') trim(str)
            write(lout1,'(<(nvar)>(x,g15.8))') data1( i, 1:nvar )
        enddo
        close(lin)
        close(lout)
        close(lout1)
        deallocate(data_, data1)      
!
! Close out
!
        write(*,*)
        write(*,'(/,a,a,a,/)') ' PPMT ',version,' FINISHED'
        stop
        
 97     stop ' ERROR allocation failed'
 98     stop ' ERROR in parameter file'
 99     stop ' ERROR in data file'
    end program ppmt

    subroutine makepar()
        open(99,file='ppmt.par',status='UNKNOWN')
        write(99,10)
    10  format( '                         Parameters for PPMT',/, &
                '                         *******************',/, &
                'START OF PARAMETERS:',/, &
                'data.dat             -input data file',/, &
                '2 3 4 0              -  number of variables, variable cols, and wt col',/, &
                '-5 1.0e7             -  trimming limits',/, &
                '25 50 50             -min/max iterations and targeted Gauss perc. (see Note 1)',/, &
                '1                    -spatial decorrelation? (0=no,1=yes) (see Note 2)',/, &
                '1 2 0                -  x, y, z columns (0=none for z)',/, &
                '50 25                -  lag distance, lag tolerance',/, &
                'nscore.out           -output data file with normal score transformed variables',/, & 
                'ppmt.out             -output data file with PPMT transformed variables',/, & 
                'ppmt.trn             -output transformation table (binary)',/, & 
                '',/, & 
                'Note 1: Optional stopping criteria, where the projection pursuit algorithm will terminate',/, & 
                'after reaching the targetted Gaussian percentile. The input percentile range is 1 (very Gaussian)',/, & 
                'to 99 (barely Gaussian); the percentiles are calculated using random Gaussian distributions.',/, & 
                'The min/max iterations overrides the targetted Gaussian percentile.',/, & 
                '',/, & 
                'Note 2: Option to apply min/max autocorrelation factors after the projection pursuit algorithm',/, & 
                'to decorrelate the variables at the specified non-zero lag distance.')
        close(99)
        return
    end subroutine makepar