!
! Module for reading and writing GSLIB-style data in binary format
! Author: Ryan M. Barnett, 2014
!
! Module Dependencies: filehandling
!   
! 
module gslib_binary
    use filehandling, only : file_open
    implicit none

!
! Public routines and variables
!
    public :: write_bin_header, & ! Write a binary header
              read_bin_header, & ! Read a binary header
              write_bin_data, & ! Write a binary header
              read_bin_data, &  ! Read a binary header
              checkgsb, &  ! Check if a file has a GSB extension
              datablock, &
              undatablock 

    ! Interface for converting different kinds of gsbv data to a single matrix array of a common kind
    ! May be useful for accumulating data for operations after reading in
    interface datablock
        module procedure datablock_i2, &
                         datablock_r4, &
                         datablock_r8
    end interface datablock

    ! Interface for converting a single matrix array into different kinds of gsbv data arrays
    ! May be useful for distributing data into their required gsbv data arrays prior to writing out
    interface undatablock
        module procedure undatablock_i2, &
                         undatablock_r4, &
                         undatablock_r8
    end interface undatablock

    type gsb_info
        character*512 header ! Arbitrary text information about the file
        integer*2 :: nvar ! Number of variables in the file
        character*64 :: vnames(50) ! Names of the variables
        integer*2 :: fmts(50) ! Formats of the variables
        integer*2 :: ndim ! Number of dimensions for all variables
        integer*4 :: dims(5) ! Size of each dimension
        integer*4 :: nrec ! Total number of records (equals multiplication of dimensions)
        integer*2 :: ikeyout ! Is a keyout array used?
        real*4 :: tmin ! Minimum trimming limits to use for masking
        real*4 :: tmax ! Minimum trimming limits to use for masking
        integer*2 :: tvar ! Variable to use for trimming/masking
        integer*4 :: nreal ! Number of realizations
        integer*2 :: ikeyout_real ! Is a unique keyout used for each realization?
        integer*4 :: nkeyout ! Length of the keyout array 
        integer*4 :: nko_index ! Length of the keyout index
        logical,  pointer :: mask(:) => null()
        integer*4, pointer :: keyout(:) => null() ! Keyout run length array 
        integer*4, pointer :: ko_index(:) => null() ! Index of non-null data_ values
    end type gsb_info

    type gsb_var
        integer*4 :: nrle ! Length of the rle arrays
        integer*4, pointer :: rle(:) => null() ! Run length encoding runs
        integer*4, pointer :: rle_index(:) => null() ! Index of data_ value for each run
        integer*2, pointer :: data_i2(:) => null() 
        real*4, pointer :: data_r4(:) => null() 
        real*8, pointer :: data_r8(:) => null()
    end type gsb_var
    
    public :: gsb_info, gsb_var

    private

    ! Temporary storage arrays and pointers
    logical, allocatable, target, dimension(:) :: ltemp11 
    logical, pointer, dimension(:) :: lpoint11
    integer*2, allocatable, target, dimension(:) :: i2temp11, i2temp12, i2temp13, i2temp14 
    integer*2, pointer, dimension(:) :: i2point11, i2point12, i2point13, i2point14
    real*4, allocatable, target, dimension(:) :: r4temp11, r4temp12
    real*4, pointer, dimension(:) :: r4point11, r4point12
    real*8, allocatable, target, dimension(:) :: r8temp11, r8temp12
    real*8, pointer, dimension(:) :: r8point11, r8point12

    contains
    
    !---------------------------------------------------------------------------
    !
    ! Routines for compressing and uncompressing a runlength encoded keyout array
    !
    !---------------------------------------------------------------------------
   

    subroutine calc_keyout( mask, nkeyout, keyout, nko_index, ko_index)
        ! Calculate a masking keyout vector based on whether
        ! the associated data value is trimmed (0) or not (1)
        implicit none
        logical, pointer, intent(in) :: mask(:)
        integer*4, pointer, intent(out) :: keyout(:), ko_index(:)
        integer*4 :: nrec, nkeyout, nko_index, irec, nrun

        ! The first value indicates whether the start of the sequence
        ! begins with a null value (0) or not (1). Null values are 
        ! not actually stored/written/read.
        nrec = ubound(mask,1)
        irec = 0; nkeyout = 1
        if( mask(1) )then
            keyout(nkeyout) = 1
        else
            keyout(nkeyout) = 0
            ! The main do loop must be initialized on a 
            ! non-null value so move through this run of
            ! null values
            irec = irec + 1
            nkeyout = nkeyout + 1
            nrun = 1
            if( irec+1  > nrec )then
                keyout(nkeyout) = nrun 
                write(*,'(a)') 'Warning - all data are null values!'
            else
                do while( not( mask(irec+1) ) )
                    irec = irec + 1
                    nrun = nrun + 1
                    if( irec == nrec ) exit  
                enddo
                keyout(nkeyout) = nrun 
            endif
        endif
        nko_index = 0
        do while( irec < nrec )
            ! Run of non-null values
            irec = irec + 1
            nkeyout = nkeyout + 1
            nko_index = nko_index + 1
            ko_index(nko_index) = irec
            nrun = 1
            if( irec+1  > nrec )then
                keyout(nkeyout) = nrun 
                exit
            else
                do while( mask(irec+1) )
                    irec = irec + 1
                    nrun = nrun + 1
                    nko_index = nko_index + 1
                    ko_index(nko_index) = irec ! Stores the index to the original data file
                    if( irec == nrec ) exit
                enddo
                keyout(nkeyout) = nrun
            endif

            if( irec+1  > nrec ) exit
            
            ! Run of null values
            irec = irec + 1
            nkeyout = nkeyout + 1
            nrun = 1
            if( irec+1  > nrec )then
                keyout(nkeyout) = nrun 
            else
                do while( not( mask(irec+1) ) )
                    irec = irec + 1
                    nrun = nrun + 1
                    if( irec == nrec ) exit
                enddo
                keyout(nkeyout) = nrun 
            endif
         enddo
         
    end subroutine calc_keyout
    
    subroutine calc_nko_index( nkeyout, keyout, nko_index )
        ! Calculate a masking keyout vector based on whether
        ! the associated data value is trimmed (0) or not (1)
        implicit none
        integer*4, intent(in) :: nkeyout
        integer*4, pointer, intent(in) :: keyout(:)
        integer*4, intent(out) :: nko_index
        integer*4 :: ikeyout
        ! The first value indicates whether the start of the sequence
        ! begins with a null value (0) or not (1). Null values are 
        ! not actually stored/written/read.

        ikeyout = 1
        if( keyout(ikeyout) == 0 )then
            ! First values are null, so initialize
            ! as such to facilitate subsequent loop
            ikeyout = ikeyout + 1
        endif
        nko_index = 0
        do while( ikeyout < nkeyout ) 
            ikeyout = ikeyout + 1
            nko_index = nko_index + keyout(ikeyout)
            if( ikeyout == nkeyout ) exit
            ikeyout = ikeyout + 1
        enddo
    end subroutine calc_nko_index

    subroutine calc_keyout_inv_i2( valsin, nkeyout, keyout, valsout )
        ! Calculate a masking keyout vector based on whether
        ! the associated data value is trimmed (0) or not (1)
        implicit none
        integer*2, pointer, intent(in) :: valsin(:)
        integer*4, intent(in) :: nkeyout
        integer*4, pointer, intent(in) :: keyout(:)
        integer*2, pointer, intent(out) :: valsout(:)
        integer*4 :: nrec, irec, nko_index, ikeyout, iko_index

        ! The first value indicates whether the start of the sequence
        ! begins with a null value (0) or not (1). Null values are 
        ! not actually stored/written/read.
        ikeyout = 1
        irec = 0
        if( keyout(ikeyout) == 0 )then
            ! First values are null, so initialize
            ! as such to facilitate subsequent loop
            ikeyout = ikeyout + 1
            nrec = irec + keyout(ikeyout)
            valsout(irec+1:nrec) = -999.99
            irec = nrec
        endif
        iko_index = 0
        do while( ikeyout < nkeyout ) 
            ikeyout = ikeyout + 1
            nko_index = iko_index + keyout(ikeyout)
            nrec = irec + keyout(ikeyout)
            valsout(irec+1:nrec) = valsin(iko_index+1:nko_index)
            iko_index = nko_index
            irec = nrec
            if( ikeyout == nkeyout ) exit
            ikeyout = ikeyout + 1
            nrec = irec + keyout(ikeyout)
            valsout(irec+1:nrec) = -999.99
            irec = nrec
        enddo
    end subroutine calc_keyout_inv_i2

    subroutine calc_keyout_inv_r4( valsin, nkeyout, keyout, valsout )
        ! Calculate a masking keyout vector based on whether
        ! the associated data value is trimmed (0) or not (1)
        implicit none
        real*4, pointer, intent(in) :: valsin(:)
        integer*4, intent(in) :: nkeyout
        integer*4, pointer, intent(in) :: keyout(:)
        real*4, pointer, intent(out) :: valsout(:)
        integer*4 :: nrec, irec, nko_index, ikeyout, iko_index

        ! The first value indicates whether the start of the sequence
        ! begins with a null value (0) or not (1). Null values are 
        ! not actually stored/written/read.
        ikeyout = 1
        irec = 0
        if( keyout(ikeyout) == 0 )then
            ! First values are null, so initialize
            ! as such to facilitate subsequent loop
            ikeyout = ikeyout + 1
            nrec = irec + keyout(ikeyout)
            valsout(irec+1:nrec) = -999.99
            irec = nrec
        endif
        iko_index = 0
        do while( ikeyout < nkeyout ) 
            ikeyout = ikeyout + 1
            nko_index = iko_index + keyout(ikeyout)
            nrec = irec + keyout(ikeyout)
            valsout(irec+1:nrec) = valsin(iko_index+1:nko_index)
            iko_index = nko_index
            irec = nrec
            if( ikeyout == nkeyout ) exit
            ikeyout = ikeyout + 1
            nrec = irec + keyout(ikeyout)
            valsout(irec+1:nrec) = -999.99
            irec = nrec
        enddo
    end subroutine calc_keyout_inv_r4
    
    subroutine calc_keyout_inv_r8( valsin, nkeyout, keyout, valsout )
        ! Calculate a masking keyout vector based on whether
        ! the associated data value is trimmed (0) or not (1)
        implicit none
        real*8, pointer, intent(in) :: valsin(:)
        integer*4, intent(in) :: nkeyout
        integer*4, pointer, intent(in) :: keyout(:)
        real*8, pointer, intent(out) :: valsout(:)
        integer*4 :: nrec, irec, nko_index, ikeyout, iko_index
        
        ! The first value indicates whether the start of the sequence
        ! begins with a null value (0) or not (1). Null values are 
        ! not actually stored/written/read.
        ikeyout = 1
        irec = 0
        if( keyout(ikeyout) == 0 )then
            ! First values are null, so initialize
            ! as such to facilitate subsequent loop
            ikeyout = ikeyout + 1
            nrec = irec + keyout(ikeyout)
            valsout(irec+1:nrec) = -999.99
            irec = nrec
        endif
        iko_index = 0
        do while( ikeyout < nkeyout ) 
            ikeyout = ikeyout + 1
            nko_index = iko_index + keyout(ikeyout)
            nrec = irec + keyout(ikeyout)
            valsout(irec+1:nrec) = valsin(iko_index+1:nko_index)
            iko_index = nko_index
            irec = nrec
            if( ikeyout == nkeyout ) exit
            ikeyout = ikeyout + 1
            nrec = irec + keyout(ikeyout)
            valsout(irec+1:nrec) = -999.99
            irec = nrec
        enddo
    end subroutine calc_keyout_inv_r8
  
    !---------------------------------------------------------------------
    !
    ! Routines for compressing and uncompressing a runlength encoded array
    !
    !---------------------------------------------------------------------

    subroutine calc_rle_keyout( vals, keyout, nko_index, ko_index,&
                                               &  nrle, rle, rle_index )
        ! Calculate the run length encoding (RLE) of an input
        ! integer array. The RLE is performed as a second level 
        ! of compression beyond the keyout (which masks out null values)
        implicit none
        integer*2, pointer, intent(in) :: vals(:)
        integer*4, pointer, intent(in) :: keyout(:), ko_index(:)
        integer*4, pointer, intent(out) :: rle(:), rle_index(:)
        integer*4 :: iko_index, nko_index, nrle, nrun, ikeyout

        ! Initialize keyout 
        if( keyout(1) )then
            ! Non-null value in first record so initialize here
            ikeyout = 1
        else
            ! Null value in first record so initialize in second
            ikeyout = 2
        endif  

        ! Determine the modified run length encoding
        nrle = 0; iko_index = 0
        do while( iko_index < nko_index )
            ! Query the current keyout run of non-null values
            ikeyout = ikeyout + 1
            iko_index = iko_index + 1
            nrle = nrle + 1
            rle_index(nrle) = ko_index(iko_index)
            nrun = 1
            if( iko_index== nko_index )then
                ! Record the run of this integer
                rle(nrle) = nrun
            else
                do while( vals(ko_index(iko_index)) == vals(ko_index(iko_index+1)) )
                    iko_index = iko_index + 1
                    nrun = nrun + 1
                    if( iko_index == nko_index ) exit
                enddo
                rle(nrle) = nrun
                ! Move forward through the non-null values
                ikeyout = ikeyout + 1
            endif
         enddo
    end subroutine calc_rle_keyout
    
    subroutine calc_rle( vals, nrle, rle, rle_index)
        ! Calculate the run length encoding (RLE) of an input
        ! integer array.
        implicit none
        integer*2, pointer, intent(in) :: vals(:)
        integer*4, pointer, intent(out) :: rle(:), rle_index(:)
        integer*4 :: irec, nrec, nrle, nrun

        ! Determine the modified run length encoding
        irec = 0 ! Index  of the original records
        nrec = ubound(vals,1)
        nrle = 0 ! Index of the number of of runs
        do while( irec < nrec )
            irec = irec + 1 
            nrle = nrle + 1
            ! Record the index of this integer
            rle_index(nrle) = irec
            nrun = 1 ! Length of the run
            if( irec == nrec )then
                ! Record the run of this integer
                rle(nrle) = nrun
            else
                do while( vals(irec) == vals(irec+1) )
                    irec = irec + 1
                    nrun = nrun + 1
                    if( irec == nrec ) exit
                enddo
                rle(nrle) = nrun
            endif
         enddo
    end subroutine calc_rle 
    
    subroutine calc_rle_inv( valsin, nrle, rle, valsout )
        ! Calculate the inverse of run length encoding (RLE), 
        ! yielding the uncompressed values.
        implicit none
        integer*2, pointer, intent(in) :: valsin(:)
        integer*4, pointer, intent(in) :: rle(:)
        integer*2, pointer, intent(out) :: valsout(:)
        integer*4 :: irec, nrec, nrle, irle
        irec = 0 ! Index of the uncompressed data array
        do irle = 1,nrle ! Index of the compressed data array
            nrec = irec + rle(irle)
            valsout(irec+1:nrec) = valsin(irle)
            irec = nrec
        enddo
    end subroutine calc_rle_inv
    
    
    !---------------------------------------------------------------------
    !
    ! Block of routines for writing a binary file
    !
    !---------------------------------------------------------------------
    
    subroutine write_bin_header( fid, fname, gsb, gsbv )
        ! Required input                                
        character*512, intent(in)  :: fname
        type(gsb_info), intent(inout) :: gsb
        type(gsb_var), pointer, intent(inout) :: gsbv(:)
        integer, intent(out) :: fid
        integer*4 len, i, test

        ! Open the output file with stream access
        fid = file_open( fname, .true., 'stream' )
        if( fid < 0 ) i = error_report(2) ! Stop
        len = len_trim(gsb.header)
        write(fid) len 
        write(fid) gsb.header(1:len) ! Header info (char)
        write(fid) gsb.nvar ! Number of variables
        write(fid) gsb.ikeyout ! Keyout array?
        write(fid) gsb.ndim 
        write(fid) gsb.dims(1:gsb.ndim)
        write(fid) gsb.nreal 
        if( gsb.ikeyout .and. gsb.nreal > 1 )write(fid) gsb.ikeyout_real
        do i = 1,gsb.nvar ! Write information for each variable
            len = len_trim(gsb.vnames(i)) 
            write(fid) len
            write(fid) gsb.vnames(i)(1:len) ! Variable name (char)
            write(fid) gsb.fmts(i) ! Type code (1=i2,2=r4,3=r8)
        enddo
        gsb.nrec = 1
        do i = 1,gsb.ndim
            gsb.nrec = gsb.nrec*gsb.dims(i)
        enddo
        
        ! Allocate the required arrays
        test = 0
        if( gsb.ikeyout )then
            allocate( gsb.mask(gsb.nrec), gsb.keyout(gsb.nrec), &
                                  & gsb.ko_index(gsb.nrec), stat = test ) 
        endif
        
        allocate( gsbv(gsb.nvar), stat = test )
        do i = 1,gsb.nvar
            if( gsb.fmts(i) == 1 )then
               allocate(gsbv(i).data_i2(gsb.nrec),gsbv(i).rle(gsb.nrec),&
                         & gsbv(i).rle_index(gsb.nrec), stat = test )
            elseif( gsb.fmts(i) == 2 )then 
                allocate( gsbv(i).data_r4(gsb.nrec), stat = test )
            else
                allocate( gsbv(i).data_r8(gsb.nrec), stat = test )           
            endif
        enddo
        if( test /= 0 ) i = error_report(1)
        
    end subroutine write_bin_header
                           
    subroutine write_bin_data( fid, gsb, gsbv, ireal )
        type(gsb_info), intent(inout) :: gsb
        type(gsb_var), pointer, intent(inout) :: gsbv(:)
        integer, intent(in) :: fid
        integer*4, intent(in) :: ireal
        integer*4 i, size_real, n
        integer*4, allocatable, dimension(:) :: size_var

        !
        ! Prepare to write the data
        !
        
        size_real = 0 ! Tracks the size of this realization
           
        allocate( size_var(gsb.nvar) )
        ! If required, calculate the keyout array
        if( gsb.ikeyout )then
            if( ireal == 1 .or. gsb.ikeyout_real )then
                if( gsb.fmts(gsb.tvar) == 1 )then
                    gsb.mask = logical( gsbv(gsb.tvar).data_i2 >= gsb.tmin .and. gsbv(gsb.tvar).data_i2 <= gsb.tmax )
                elseif( gsb.fmts(gsb.tvar) == 2 )then
                    gsb.mask = logical( gsbv(gsb.tvar).data_r4 >= gsb.tmin .and. &
                                        & gsbv(gsb.tvar).data_r4 <= gsb.tmax )
                else
                    gsb.mask = logical( gsbv(gsb.tvar).data_r8 >= gsb.tmin .and. &
                                        & gsbv(gsb.tvar).data_r8 <= gsb.tmax )
                endif
                ! Apply the mask to generate a keyout array
                ! and values with null values removed
                call calc_keyout( gsb.mask, gsb.nkeyout, gsb.keyout, gsb.nko_index, gsb.ko_index )
                ! Position = (nkeyout + keyout)*int4
                size_real = size_real + (1 + gsb.nkeyout)*4
            endif
        endif  
         
        do i = 1,gsb.nvar
            if( gsb.fmts(i) > 1 )then
                ! Real values - no run length encoding (though 
                ! potentially a keyout array)
                if( gsb.ikeyout )then
                    n = gsb.nko_index
                else
                    n = gsb.nrec
                endif
                ! Position = (vals)*precision
                if( gsb.fmts(i) == 2 )then
                    size_var(i) = n*4 !r4
                else
                    size_var(i) = n*8 ! r8
                endif    
            else
                if( gsb.ikeyout )then
                    call calc_rle_keyout( gsbv(i).data_i2, gsb.keyout, &
                         &gsb.nko_index, gsb.ko_index, n,gsbv(i).rle, gsbv(i).rle_index )
                                                                   
                else
                    call calc_rle( gsbv(i).data_i2, n, gsbv(i).rle, gsbv(i).rle_index )
                endif
                gsbv(i).nrle = n ! Store the length of the rle sequence
                ! Size = (nrle + rle + rle_vals)*int4
                size_var(i) = (1 + n)*4 + n*2
            endif
            size_real = size_real + size_var(i) + 4
        enddo
                       
        !
        ! The information that is needed for writing has now been stored. Now, actually write it.
        !
        
        write(fid) size_real 
        if( gsb.ikeyout )then
            if( ireal == 1 .or. gsb.ikeyout_real )then
                write(fid) gsb.nkeyout
                write(fid) gsb.keyout(1:gsb.nkeyout)
            endif
        endif
        do i = 1,gsb.nvar
            write(fid) size_var(i)
            if( gsb.fmts(i) >  1 )then
                if( not(gsb.ikeyout) )then
                    if( gsb.fmts(i) ==  2 )then
                        write(fid) gsbv(i).data_r4(1:gsb.nrec)
                    else
                        write(fid) gsbv(i).data_r8(1:gsb.nrec)
                    endif
                else
                    if( gsb.fmts(i) ==  2 )then
                        write(fid) gsbv(i).data_r4( gsb.ko_index( 1:gsb.nko_index ) ) 
                    else
                        write(fid) gsbv(i).data_r8( gsb.ko_index( 1:gsb.nko_index ) )
                    endif
                endif
            else
                write(fid) gsbv(i).nrle
                write(fid) gsbv(i).rle( 1:gsbv(i).nrle )
                write(fid) gsbv(i).data_i2( gsbv(i).rle_index( 1:gsbv(i).nrle ) )
            endif
        enddo
    endsubroutine write_bin_data
    
    !---------------------------------------------------------------------
    !
    ! Block of routines for reading a binary file
    !
    !---------------------------------------------------------------------
    subroutine read_bin_header( fid, fname, gsb, gsbv )
        ! Required input                                
        character*512,intent(in)  :: fname
        integer, intent(out) :: fid
        type(gsb_info), intent(out)  :: gsb
        type(gsb_var), pointer, intent(out)  :: gsbv(:)
        integer*4 len, i, test

        
        ! Open the input data file with stream access
        fid = file_open( fname, .false., 'stream' )
        if( fid < 0 ) i = error_report(3)
        
        read(fid) len 
        read(fid) gsb.header(1:len) ! Header info (char)
        read(fid) gsb.nvar ! Number of variables
        read(fid) gsb.ikeyout ! Keyout array?
        read(fid) gsb.ndim
        read(fid) gsb.dims(1:gsb.ndim) ! Dimensions
        read(fid) gsb.nreal
        if( gsb.nreal > 1 )then
            if( gsb.ikeyout ) read(fid) gsb.ikeyout_real        
        endif
        do i = 1,gsb.nvar 
            read(fid) len
            read(fid) gsb.vnames(i)(1:len) ! Variable name (char)
            read(fid) gsb.fmts(i) ! Type code (1=i2,2=r4,3=r8)
        enddo
        gsb.nrec = 1
        do i = 1,gsb.ndim
            gsb.nrec = gsb.nrec*gsb.dims(i)
        enddo
        
        ! Allocate the required arrays
        test = 0
        if( gsb.ikeyout )then
            allocate( gsb.mask(gsb.nrec), gsb.keyout(gsb.nrec), &
                                  & gsb.ko_index(gsb.nrec), stat = test ) 
        endif
        allocate( gsbv(gsb.nvar), stat = test )
        do i = 1,gsb.nvar
            if( gsb.fmts(i) == 1 )then
               allocate(gsbv(i).data_i2(gsb.nrec),gsbv(i).rle(gsb.nrec),&
                         & gsbv(i).rle_index(gsb.nrec), stat = test )
            elseif( gsb.fmts(i) == 2 )then 
                allocate( gsbv(i).data_r4(gsb.nrec), stat = test )
            else
                allocate( gsbv(i).data_r8(gsb.nrec), stat = test )           
            endif
        enddo

        if( test /= 0 ) i = error_report(1)
        
    end subroutine read_bin_header
     
    subroutine skip_pos( fid, size, test )
        integer, intent(in) :: fid
        integer*4, intent(in) :: size
        integer*8 :: mypos
        integer*1 :: junk
        integer test
        read(fid) junk
        inquire( fid, POS = mypos)
        mypos = mypos + size - 1
        test = 0
        read( fid,  POS = mypos, end = 8 )
        return
8       test = 1
    end subroutine skip_pos
    
    subroutine current_pos( fid, mypos )
        integer, intent(in) :: fid
        integer*8, intent(out) :: mypos
        integer*1 :: junk
        write(fid) junk
        inquire( fid, POS = mypos)
        mypos = mypos - 1
    end subroutine current_pos
    
    subroutine read_bin_keyout( fid, gsb )
        integer, intent(in) :: fid
        type(gsb_info), intent(inout) :: gsb
        
        read(fid) gsb.nkeyout
        read(fid) gsb.keyout(1:gsb.nkeyout)
        call calc_nko_index( gsb.nkeyout, gsb.keyout, gsb.nko_index )
    endsubroutine read_bin_keyout

    subroutine read_bin_var( fid, gsb, gsbv, ivar  )
        integer, intent(in) :: fid
        integer*2, intent (in) :: ivar
        type(gsb_info), intent(inout) :: gsb
        type(gsb_var), pointer, intent(inout) :: gsbv(:)
        integer*4 n
        if( gsb.fmts(ivar) >  1 )then
            if( not(gsb.ikeyout) )then
                if( gsb.fmts(ivar) == 2 )then
                    read(fid) gsbv(ivar).data_r4
                else
                    read(fid) gsbv(ivar).data_r8
                endif
            else
                if( gsb.fmts(ivar) == 2 )then
                    allocate( r4temp11(gsb.nko_index) )
                    r4point11 => r4temp11
                    read(fid) r4point11 ! Temporary storage of input compressed keyout data
                    call calc_keyout_inv_r4( r4point11, gsb.nkeyout, gsb.keyout, gsbv(ivar).data_r4 )
                    nullify(r4point11); deallocate(r4temp11)  
                 else
                    allocate( r8temp11(gsb.nko_index) )
                    r8point11 => r8temp11
                    read(fid) r8point11 ! Temporary storage of input compressed keyout data
                    call calc_keyout_inv_r8( r8point11, gsb.nkeyout, gsb.keyout, gsbv(ivar).data_r8 )
                    nullify(r8point11); deallocate(r8temp11)  
                 endif
            endif
        else
            read(fid) gsbv(ivar).nrle
            read(fid) gsbv(ivar).rle( 1:gsbv(ivar).nrle )
            n = sum( gsbv(ivar).rle( 1:gsbv(ivar).nrle ) )
            allocate( i2temp11(gsbv(ivar).nrle), i2temp12(n) )
            i2point11 => i2temp11 ! Temporary storage of input compressed rle data
            i2point12 => i2temp12 ! Temporary storage of output uncompressed rle data
            read(fid) i2point11 
            call calc_rle_inv( i2point11, gsbv(ivar).nrle, gsbv(ivar).rle, i2point12 )
            nullify(i2point11); deallocate(i2temp11)  
            if( not(gsb.ikeyout) )then
                gsbv(ivar).data_i2 = i2point12
            else
                call calc_keyout_inv_i2( i2point12, gsb.nkeyout, gsb.keyout, gsbv(ivar).data_i2 )
            endif
            nullify(i2point12); deallocate(i2temp12) 
        endif
    endsubroutine read_bin_var
    
    subroutine read_bin_data( fid, gsb, gsbv, nvari, vcols, ireal, freal )
        integer, intent(in) :: fid, nvari, ireal, freal, vcols(nvari)
        type(gsb_info), intent(inout) :: gsb
        type(gsb_var), pointer, intent(inout) :: gsbv(:)
        integer*4 :: size
        integer icol, test, i

        read(fid) size
        if( ireal < freal)then
            if( ireal == 1 .and. gsb.ikeyout .and. gsb.ikeyout_real < 1 )then
                ! Have to read and store the keyout
                call read_bin_keyout( fid, gsb  )
                size = size - (gsb.nkeyout + 1)*4
            endif
            ! Skip the remainder of this realization
            call skip_pos( fid, size, test ) 
            if( test /= 0 )then; return; endif
            return
            
        elseif( ireal == 1 .and. gsb.ikeyout )then
            call read_bin_keyout( fid, gsb )
        elseif( gsb.ikeyout_real > 0 )then
            call read_bin_keyout( fid, gsb )   
        endif
        
        icol = 1
        do i = 1,gsb.nvar
            !call current_pos( fid, mypos )
            read(fid) size
            if( icol > nvari )then
                call skip_pos( fid, size, test ) 
                if( test /= 0 )then; return; endif
                cycle  
            elseif( i /= vcols(icol) )then
                call skip_pos( fid, size, test ) 
                if( test /= 0 )then; return; endif
                cycle
            endif
            
            ! Read in this variable
            call read_bin_var( fid, gsb,gsbv, vcols(icol)  )
            icol = icol + 1
        enddo
    end subroutine read_bin_data
        
    ! Test if the file extension is .gsb
    logical function checkGSB( fname ) result ( igsb )
        character*(*), intent (in) :: fname
        character*4 :: test
        integer :: los
        igsb = .false.
        los = len_trim( fname )
        test = fname(los-3:los)
        if( test == '.gsb' )then
            igsb = .true.
        elseif( test == '.GSB' )then 
            igsb = .true.
        endif
    end function checkGSB

    !--------------------------------------------------------------------
    ! Subroutine for accumulating from potentially different gsbv types
    ! (of potentially different kinds) and entering them into a 'datablock',
    ! which is an array of integer*4 (but drawn from gsbv.data_i2, 
    ! gsbv.data_r4, gsbv_r8)
    ! R. Barnett - 2015
    !--------------------------------------------------------------------
    subroutine datablock_i2( gsb, gsbv, vcols, data_  )
        implicit none
        type(gsb_info), intent(in) :: gsb
        type(gsb_var), pointer, intent(in) :: gsbv(:)
        integer, intent(in) :: vcols(:)
        integer*2, intent(out) :: data_(:,:)
        integer i, ndata, nvar, nvari

        ndata = gsb.nrec
        nvari = gsb.nvar
        nvar = size(vcols)
        if( size(data_,2) /= ndata ) i = error_report(4)  
        ! Create the data block
        do i = 1, nvar
            if( gsb.fmts(vcols(i)) == 1 )then
                data_(i,:) = gsbv(vcols(i)).data_i2
            elseif( gsb.fmts(vcols(i)) == 2 )then
                data_(i,:) = nint(gsbv(vcols(i)).data_r4)
            else
                data_(i,:) = nint(gsbv(vcols(i)).data_r8)
            endif
        enddo
    end subroutine datablock_i2

    !--------------------------------------------------------------------
    ! Subroutine for accumulating from potentially different gsbv types
    ! (of potentially different kinds) and entering them into a 'datablock',
    ! which is an array of real*4 (but drawn from gsbv.data_i2, 
    ! gsbv.data_r4, gsbv_r8)
    ! R. Barnett - 2015
    !--------------------------------------------------------------------
    subroutine datablock_r4( gsb, gsbv, vcols, data_  )
        implicit none
        type(gsb_info), intent(in) :: gsb
        type(gsb_var), pointer, intent(in) :: gsbv(:)
        integer, intent(in) :: vcols(:)
        real*4, intent(out) :: data_(:,:)
        integer i, ndata, nvar, nvari

        ndata = gsb.nrec
        nvari = gsb.nvar
        nvar = size(vcols)
        if( size(data_,2) /= ndata ) i = error_report(4)
        ! Create the data block
        do i = 1, nvar
            if( gsb.fmts(vcols(i)) == 1 )then
                data_(i,:) = real(gsbv(vcols(i)).data_i2)
            elseif( gsb.fmts(vcols(i)) == 2 )then
                data_(i,:) = gsbv(vcols(i)).data_r4
            else
                data_(i,:) = gsbv(vcols(i)).data_r8
            endif
        enddo
    end subroutine datablock_r4

    !--------------------------------------------------------------------
    ! Subroutine for accumulating from potentially different gsbv types
    ! (of potentially different kinds) and entering them into a 'datablock',
    ! which is an array of real*8 (but drawn from gsbv.data_i2, 
    ! gsbv.data_r4, gsbv_r8)
    ! R. Barnett - 2015
    !--------------------------------------------------------------------
    subroutine datablock_r8( gsb, gsbv, vcols, data_  )
        implicit none
        type(gsb_info), intent(in) :: gsb
        type(gsb_var), pointer, intent(in) :: gsbv(:)
        integer, intent(in) :: vcols(:)
        real*8, intent(out) :: data_(:,:)
        integer i, ndata, nvar, nvari

        ndata = gsb.nrec
        nvari = gsb.nvar
        nvar = size(vcols)
        if( size(data_,2) /= ndata ) i = error_report(4)   
        ! Create the data block
        do i = 1, nvar
            if( gsb.fmts(vcols(i)) == 1 )then
                data_(i,:) = dble(gsbv(vcols(i)).data_i2)
            elseif( gsb.fmts(vcols(i)) == 2 )then
                data_(i,:) = dble(gsbv(vcols(i)).data_r4)
            else
                data_(i,:) = gsbv(vcols(i)).data_r8
            endif
        enddo
    end subroutine datablock_r8

    !--------------------------------------------------------------------
    ! Subroutine for 'un'accumulating a datablock array into potentially 
    ! different gsbv kinds (preparing to write out to binary). This subroutine
    ! operates on an input integer*2 array.
    ! R. Barnett - 2015
    !--------------------------------------------------------------------
    subroutine undatablock_i2( gsb, vcols, data_, gsbv  )
        implicit none
        type(gsb_info), intent(in) :: gsb
        integer, intent(in) :: vcols(:)
        integer*2, intent(in) :: data_(:,:)
        type(gsb_var), pointer, intent(inout) :: gsbv(:)
        integer i, ndata, nvar, nvari

        ndata = gsb.nrec
        nvari = gsb.nvar
        nvar = size(vcols)
        if( size(data_,2) /= ndata ) i = error_report(4)
        ! Divide the datablock into the appropriate GSBV kinds
        do i = 1, nvar
            if( gsb.fmts(vcols(i)) == 1 )then
                gsbv(vcols(i)).data_i2 = data_(i,:)
            elseif( gsb.fmts(vcols(i)) == 2 )then
                gsbv(vcols(i)).data_r4 = real(data_(i,:))
            else
                gsbv(vcols(i)).data_r8 = dble(data_(i,:))
            endif
        enddo
    end subroutine undatablock_i2


    !--------------------------------------------------------------------
    ! Subroutine for 'un'accumulating a datablock array into potentially 
    ! different gsbv kinds (preparing to write out to binary). This subroutine
    ! operates on an input real*4 array.
    ! R. Barnett - 2015
    !--------------------------------------------------------------------
    subroutine undatablock_r4( gsb, vcols, data_, gsbv  )
        implicit none
        type(gsb_info), intent(in) :: gsb
        integer, intent(in) :: vcols(:)
        real*4, intent(in) :: data_(:,:)
        type(gsb_var), pointer, intent(inout) :: gsbv(:)
        integer i, ndata, nvar, nvari

        ndata = gsb.nrec
        nvari = gsb.nvar
        nvar = size(vcols)
        if( size(data_,2) /= ndata ) i = error_report(4)   
        ! Divide the datablock into the appropriate GSBV kinds
        do i = 1, nvar
            if( gsb.fmts(vcols(i)) == 1 )then
                gsbv(vcols(i)).data_i2 = nint(data_(i,:))
            elseif( gsb.fmts(vcols(i)) == 2 )then
                gsbv(vcols(i)).data_r4 = data_(i,:) 
            else
                gsbv(vcols(i)).data_r8 = data_(i,:) 
            endif
        enddo
        
    end subroutine undatablock_r4

    !--------------------------------------------------------------------
    ! Subroutine for 'un'accumulating a datablock array into potentially 
    ! different gsbv kinds (preparing to write out to binary). This subroutine
    ! operates on an input real*8 array.
    ! R. Barnett - 2015
    !--------------------------------------------------------------------
    subroutine undatablock_r8( gsb, vcols, data_, gsbv  )
        implicit none
        type(gsb_info), intent(in) :: gsb
        integer, intent(in) :: vcols(:)
        real*8, intent(in) :: data_(:,:)
        type(gsb_var), pointer, intent(inout) :: gsbv(:)
        integer i, ndata, nvar, nvari

        ndata = gsb.nrec
        nvari = gsb.nvar
        nvar = size(vcols)
        if( size(data_,2) /= ndata ) i = error_report(4)
        ! Divide the datablock into the appropriate GSBV kinds
        do i = 1, nvar
            if( gsb.fmts(vcols(i)) == 1 )then
                gsbv(vcols(i)).data_i2 = nint(data_(i,:))
            elseif( gsb.fmts(vcols(i)) == 2 )then
                gsbv(vcols(i)).data_r4 = data_(i,:) 
            else
                gsbv(vcols(i)).data_r8 = data_(i,:) 
            endif
        enddo
    end subroutine undatablock_r8


    integer function error_report( errcode ) result(i)
    !--------------------------------------------------------------------
    ! Subroutine for reporting the error to the user and stopping the program.
    ! R. Barnett - 2015
    !--------------------------------------------------------------------
        implicit none
        integer, intent(in) :: errcode
        select case (errcode)
            case(1) 
                write(*,'(/,a,/)') 'ERROR: allocation failed for GSB array' 
            case(2)
                write(*,'(/,a,/)') 'ERROR: could not open output GSB file'
            case(3)
                write(*,'(/,a,/)') 'ERROR: could not open input GSB file'
            case(4)
                write(*,'(/,a,/,a)') 'ERROR: second dimension of datablock array '&
                                     '       should match gsb.nrec'
        end select
        i = errcode
        stop ! Just stop everything for now
     end function error_report
     
end module gslib_binary


