!***********************************************************************************!
!                                                                                   !
!  PROGRAM: PPMT_B                                                                  !
!                                                                                   !
!  PURPOSE:  Projection Pursuit Multivariate Transform (Back-Transform)             !
!                                                                                   !
!  AUTHOR: Ryan Barnett 2014                                                        !
!***********************************************************************************!

    program ppmt_b
        use geostat ! Common variables
        use gslib_binary ! Routines for GSLIB-style binary read/write
        use filehandling ! Routines for GSLIB-style ASCII read/write and other functionality
        use backtrmod ! Various functions and subroutines   
        use trans_trend ! N(0,1) enforcement prior to back-transform routines
        use quicksort ! Sorting routines
        implicit none
        logical :: usetrans, uselocal, usegsb, usegsbout
        logical, allocatable, dimension(:) :: spatialiter
        integer*4 :: iters, np0, len, iters0, iters1, iters00, iters10, itertype
        integer*4, allocatable :: ixp(:), gsbcols(:)
        real*8 :: zmin0
        real*8, allocatable, dimension(:) :: snorm, xtemp
        real*8, allocatable, dimension(:,:) :: zd, z, x, xt, Sinv, pr, U, Ux, yd2, zd2
        real*8, allocatable, dimension(:,:,:) :: Us, Cs
        character*5, parameter :: version = '3.201'
        character*12, parameter :: parname = 'ppmt_b.par'
        ! Back-transform routine parameters
        integer*4 :: utail, ltail
        real*8 :: utpar, ltpar, pdelta, gp
        ! Histogram reproduction variables
        integer*4 :: icoll(100), nloop, nxyz, np, ireal, nreal
        character*512 :: localfl
        real*8 :: wtfac
        real*8, allocatable :: tval(:), kv(:,:), kvtemp(:)
        real*8, allocatable :: tcdf(:)       
        
        ! Custom types for GSB reading/writing
        type(gsb_info) :: gsb_in, gsb_out! General information
        type(gsb_var), pointer :: gsbv_in(:), gsbv_out(:) ! Variable information 
!
! Note version and set fixed parameters
!   
        write(*,'(/,a,a,/)') ' PPMT_B Reverse Projection version: ',version 
!
! Get the name of parameter file, open the file and find start of parameters
!
        call findpar(parname,1,lin)
!
! Read Input Parameters:
        read(lin,'(a)',err=98) datafile(1)
        call checkname(datafile(1))
        write(*,'(a,a)') ' transformation file = ',trim(datafile(1))
        
        read(lin,'(a)',err=98) datafile(2)
        call checkname(datafile(2))
        usegsb = checkGSB( datafile(2) ) ! Test GSB
        if( usegsb )then
            write(*,'(a,a)') ' GSB-style data file = ',datafile(2)(1:40)
        else
            write(*,'(a,a)') ' GSLIB-style data file = ',datafile(2)(1:40)
        endif        
        read(lin,*,err=98) nvar, vcols(1:nvar)
        write(*,'(a,i4)') ' number of variables = ', nvar
        write(*,'(a,<(nvar)>(x,i4))') ' cols. for transformed vars. = ', vcols(1:nvar)

        read(lin,*,err=98) tmin,tmax
        write(*,'(a,2(x,e7.2))') ' trimming limits = ',tmin,tmax
        
        read(lin,*,err=98) nx,ny,nz,nreal
        write(*,'(a,<(4)>(x,i8))') ' nx, ny, nz, nreal = ',nx,ny,nz,nreal
        nxyz = nx*ny*nz
        
        if( nxyz <= 0 ) write(*,'(a)') ' assuming one realization since nx*ny*nz = 0'
            
        read(lin,'(a)',err=98) outfl
        call checkname(outfl)
        write(*,'(a,a)') ' output file = ',trim(outfl)
        usegsbout = checkgsb( outfl )
        if( usegsb .and. not(usegsbout) )then
            write(*,'(/,a,/)') 'ERROR: require output GSB file if using input GSB file' 
            stop
        endif

        read(lin,*,err=98) i
        write(*,'(a,2(x,i2))') ' enforce N(0,1)? = ',i
        
        usetrans = .false.
        uselocal = .false.
        if( i > 0 )then
            usetrans = .true.
            read(lin,*,err=98);read(lin,*,err=98);read(lin,*,err=98)
            read(lin,*,err=98) i
            write(*,*) ' account for local data (1=yes) = ',i
            
            if( i > 0  )then
                uselocal = .true.
                read(lin,'(a512)',err=98) localfl 
                call checkname(localfl)
                write(*,*) ' local file = ',localfl(1:40)

                read(lin,*,err=98) icoll(1:nvar)
                write(*,*) ' columns for kriging variance = ',icoll(1:nvar)

                read(lin,*,err=98) wtfac
                write(*,*) ' scaling factor = ',wtfac
                ! Scale from 0 to 1  --> 0.33 to 3.0
                wtfac = 0.33d0 + wtfac*(3.d0-0.33d0)
                read(lin,*,err=98) nloop
                write(*,*) ' number of iterations = ',nloop
             endif
        endif
        close(lin)
        
        ! Hardcode some parameters for now
        utail = 2; utpar = 2
        ltail = 2; ltpar = 2     
!
! Determine nxyz if necessary
!
        if( nxyz <= 0 .and. not(usegsb) )then
            call datafilesize ( lin, datafile(2), nvari, nxyz )
            nreal = 1
            close( lin )
        endif           
!
! Load the transformation file information
!  
        lin = file_open( datafile(1), .false., 'stream' )
        read( lin, err = 97 ) len 
        read( lin, err = 97 ) str(1:len)
        if( str(1:len) /= 'PPMT transformation table file' ) go to 97
        read( lin, err = 97 ) i   
        if( i /= nvar )then
            write(*,'(a)') ' number of variables in the transformation file'
            write(*,'(a)') ' and parameter file does not match! '
            stop
        endif
        read( lin, err = 97 ) ndata  
        iters = 500
        allocate( zd(ndata,nvar) , z(nvar,nxyz), Sinv(nvar,nvar), snorm(ndata), vnames(nvar),&
                & pr(ndata,iters) , U(nvar,nvar),Us(nvar,nvar,iters), zd2(ndata,nvar),  &
                & yd2(ndata,nvar), Cs(nvar,nvar,iters), spatialiter(iters), stat=test )
        if( uselocal )then ! File with local kriging variances?
            allocate( kv(nxyz,nvar), stat=test )
            call krig_var( localfl, icoll(1:nvar), wtfac, kv )
        endif
        if(test /= 0)then; write(*,'(/,a,/)') 'ERROR: ALLOCATION FAILED';stop;endif
        !Read in the normal score preprocessing info
        read( lin, err = 97 ) snorm
        read( lin, err = 97 ) zd
        !Read in the inverse sphere matrix
        read( lin, err = 97 ) Sinv
        
        iters = 0
        iters0 = 0 ! Counter for spatial decorrelation
        iters1 = 0 ! Counter for projection pursuit
        spatialiter = .false.
        do 
            if( iters > 500 )then
                write(*,*)  ' PPMT back-transform only handles 500 iterations for now....'
                stop
            endif
            read( lin, err = 97 ) itertype
            if( itertype == 12345 )then
                spatialiter(iters+1) = .true.
                iters0 = iters0 + 1 ! Spatial decorrelation
                read( lin, err = 97 ) Cs(:,:,iters0)
            elseif( itertype == 54321 )then
                iters1 = iters1 + 1 ! Projection pursuit
                read( lin, err = 97 ) Us(:,:,iters1)
                read( lin, err = 97 ) pr(:,iters1)
            elseif( itertype == 6789 )then
                exit ! Final marginal normal score
            else
                go to 97 ! Shouldn't get here - error
            endif
            iters = iters + 1
        enddo
        read(lin, err = 97 ) zd2
        read(lin, err = 97 ) yd2
        do i = 1,nvar
            call qsortem( yd2(:,i), 1, ndata, zd2(:,i)  )
        enddo
        close(lin)
        
        !Determine extrapolation values
        ymin = snorm(1)  !snorm_inv ( dble(1/dble(nsim)) ) 
        ymax = snorm(ndata)   !snorm_inv ( (dble(nsim)-1.0)/dble(nsim) )
        iters00 = iters0 + 1
        iters10 = iters1 + 1 
        
!
! Prepare the input and output files
!
        if( not(usegsb) )then
            lout = file_open ( outfl, .true. )
            lin = file_open ( datafile(2), .false. )
            read(lin,'(a512)',err=99) str
            write(lout,'(a,a)') 'PPMT Back Transform: ',trim(str)
            read(lin,*,err=99) nvari
            write(lout,'(i3)') nvar
            do i = 1, nvari
                read(lin,'(a512)',err=99) str
                do j = 1, nvar
                      if(i == vcols(j)) vnames(j) = str
                enddo      
            enddo
            do i = 1, nvar
                write(lout,'(a)') 'PPMT_B:'//trim(adjustl(vnames(i)))
            enddo   
        else
            call read_bin_header( lin, datafile(2), gsb_in, gsbv_in ) 
            if( gsb_in.nreal < nreal )then
                write(*,'(a)') ' ERROR: data file has less realizations'
                write(*,'(a)') '        than the parameter file specifies!'
                stop
            elseif( gsb_in.nrec < nxyz )then
                write(*,'(a)') ' ERROR: data file has less nodes'
                write(*,'(a)') '        than the parameter file specifies!'
                stop
            endif  
            gsb_out = gsb_in
            gsb_out.header = 'PPMT_B:'//trim(gsb_out.header)
            ! Keyout based on the first variable
            gsb_out.tvar = 1
            gsb_out.tmin = tmin
            gsb_out.tmax = tmax
            call write_bin_header( lout, outfl, gsb_out, gsbv_out ) 
            nvari = gsb_in.nvar
            allocate( gsbcols(nvari), stat = test )
            do i = 1,nvari
                gsbcols(i) = i
            enddo
        endif

        ! Allocating and deallocating takes time - only change allocation if required 
        ! in subsequent iterations over the realizations
        np = nxyz 
        np0 = np
        allocate( xt(np,nvar), x(np,nvar), Ux(np,nvar), stat=test )
        if( usetrans )then
            allocate( tcdf(np), tval(np), xtemp(np), stat=test )
            call ref_cdf( np, tcdf )
            do i = 1, np
                tval(i) = snorm_inv ( tcdf(i) ) 
            enddo
            if( uselocal )then
                allocate( kvtemp(np), stat=test  )
            endif
        endif  
        allocate( data_(nvari,nxyz), ixp(nxyz), stat=test ) ! This doesn't change in the iterations
        if(test /= 0)then; write(*,'(/,a,/)') 'ERROR: ALLOCATION FAILED';stop;endif
        
        utail = 2
        utpar = 1.95
        ltail = 2
        ltpar = 1.95
        pdelta = 1.d0 /dble(ndata) - 1.d0/dble(nxyz) 
        do ireal = 1, nreal 
            write(*,'(a,i3)') ' working on realization:',ireal
            ! Read in the current realization
            if( not( usegsb ) )then
                read(lin,*,err=99) data_
            else
                call read_bin_data( lin, gsb_in, gsbv_in, nvari, gsbcols, ireal, 1 )
                call datablock( gsb_in, gsbv_in, gsbcols, data_  )
            endif
            np = 0
            do j = 1, nxyz   
                if( any( data_(vcols(1:nvar),j) < tmin ) .or. any( data_(vcols(1:nvar),j) > tmax ) ) cycle
                np = np + 1
                ixp(np) = j
            enddo
            
            if( np /= np0 )then
                ! Reallocate the arrays since the number of nulls has changed
                deallocate( xt, x, Ux  )
                allocate( xt(np,nvar), x(np,nvar), Ux(np,nvar), stat=test )
                if(test /= 0)then; write(*,'(/,a,/)') 'ERROR: ALLOCATION FAILED';stop;endif
            endif
            xt = transpose( data_( vcols(1:nvar), ixp(1:np) ) )
            
           if( usetrans )then
                ! Assemble the reference cdf and associated values ( N(0,1) )
                ! if npass has changed from the previous realization
                if( np /= np0 )then
                    deallocate( tcdf, tval, xtemp  )
                    allocate( tcdf(np), tval(np),xtemp(np), stat=test )
                    if( uselocal )then
                        deallocate( kvtemp );allocate( kvtemp(np), stat=test  )
                    endif
                    if( test /= 0 )then; write(*,'(/,a,/)') 'ERROR: ALLOCATION FAILED';stop;endif
                    call ref_cdf( np, tcdf )
                    do i = 1, np
                        tval(i) = snorm_inv ( tcdf(i) ) 
                    enddo
                endif
                ! Enforce marginal N(0,1)
                do i = 1, nvar
                    if( uselocal )then
                        kvtemp = kv( ixp(1:np), i )
                        ! Enforce local data
                        call trans( nloop, xt(:,i), tval, xtemp, kvtemp)
                    else
                        ! Do not enforce local data
                        call trans( 1, xt(:,i), tval, xtemp)
                    endif
                    xt(:,i) = xtemp
                enddo
            endif
            np0 = np
            !Back-nscore        
            do i=1,nvar
                zmin = minval( zd2(:,i) )
                zmax = maxval( zd2(:,i) )
                do j = 1, np
                    x(j,i) = backtr( xt(j,i), zd2(:,i), yd2(:,i), zmin, zmax, ltail, ltpar, utail, utpar )
                enddo
            enddo
            xt = x
            ! Proceed backward through the iterations
            iters0 = iters00
            iters1 = iters10
            do i = iters, 1, -1
                if( spatialiter(i) )then
                    iters0 = iters0 - 1
                    ! Back-transform
                    do j = 1, np
                        x(j,:) = matmul( xt(j,:), Cs(:,:,iters0) ) 
                    enddo
                else
                    iters1 = iters1 - 1
                    ! Grab the correct vector
                    U = Us(:,:,iters1)
                    ! Project the data along the vector
                    Ux = matmul( xt, U )
                    ! Back-transform the projection
                    xt(:,1) = Ux(:,1)
                    zmin = minval( pr(:,iters1) )
                    zmax = maxval( pr(:,iters1) )
                    do j = 1, np
                        Ux(j,1) = backtr( xt(j,1), pr(:,iters1), snorm, zmin, zmax, ltail, ltpar, utail, utpar )
                    enddo
                    ! Unproject - could do this in one step but likely to cause allocation errors
                    do j = 1, np
                        x(j,:) = matmul( Ux(j,:), transpose(U) ) 
                    enddo
                endif
                xt = x
            enddo
            !Back-sphere (again - do one by one
            do i = 1, np
                x(i,:) = matmul( xt(i,:), Sinv )  
            enddo
            !Back-nscore    
            xt = x        
            do i=1,nvar
                zmin = minval( zd(:,i) )
                zmax = maxval( zd(:,i) )
                do j = 1, np
                    x(j,i) = backtr( xt(j,i), zd(:,i), snorm, zmin, zmax, ltail, ltpar, utail, utpar )
                enddo
            enddo
            z = -999.999  ! Output is null by default  
            if( np > 0 ) z( :, ixp(1:np) ) = transpose(x)
            if( not(usegsb) )then
                write(lout,'(<nvar>(x,g15.8))') z    
            else
                call undatablock( gsb_out, vcols, z, gsbv_out  )
                call write_bin_data( lout, gsb_out, gsbv_out, ireal )
            endif
        enddo
        close(lin)
        close(lout)
!   
! Close out
!
        write(*,*)
        write(*,'(/,a,a,a,/)') ' PPMT_B ',version,' FINISHED'
        stop
97      stop 'ERROR in transformation table file!'
98      stop 'ERROR in parameter file!'
99      stop 'ERROR in data file!'
    end program ppmt_b

    subroutine makepar()
        open(99,file='ppmt_b.par',status='UNKNOWN')
        write(99,10)
    10  format( '                 PPMT Back Transformation',/, &
                '                 ************************',/,/, &
                'START OF PARAMETERS:',/, & 
                'ppmt.trn                      -file with transformation data',/, & 
                'sgsim.out                     -file with data to back-transform (GSB detection)',/, & 
                '2 1 2                         - nvar, column numbers',/, & 
                '-9.0 1.0e21                   - trimming limits',/, & 
                '250 250 1 1                   - nx, ny, nz, nreal (ignored if nx = 0)',/, & 
                'ppmt_b.out                    -file for back transformed values',/, & 
                '1                             -enforce N(0,1)? (0=no,1=yes)',/, & 
                '-----------------------------------------------------------------',/, & 
                '     Disregard the following lines if not enforcing N(0,1) ',/, & 
                '-----------------------------------------------------------------',/, & 
                '1                             -consider local uncertainty? (0=no,1=yes)',/, &        
                'kt3dv.out                     - local file',/, & 
                '1 2                           -  columns for kriging variances',/, & 
                '.2                            -  weight factor',/, &  
                '5                             -  number of loops') 
        close(99)    
        return
    end subroutine makepar
    
